# Common environment variables for client/exporter config
JMP_CLIENT_CONFIG_HOME = "JMP_CLIENT_CONFIG_HOME"
JMP_CLIENT_CONFIG = "JMP_CLIENT_CONFIG"
JMP_NAMESPACE = "JMP_NAMESPACE"
JMP_NAME = "JMP_NAME"
JMP_ENDPOINT = "JMP_ENDPOINT"
JMP_TOKEN = "JMP_TOKEN"
JMP_DRIVERS_ALLOW = "JMP_DRIVERS_ALLOW"
JUMPSTARTER_HOST = "JUMPSTARTER_HOST"
JMP_LEASE = "JMP_LEASE"
